#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#define OBJ_UNKNOWN     0
#define OBJ_BAD         1
#define OBJ_MSG         2
#define OBJ_MSG_END     3
#define OBJ_MSG_BEGIN   4
#define OBJ_MSG_CORRUPT 5

struct messageblock
{
  int marker;
  int twohundred;
  int msglen;
  int nextblock;
  char buffer[0x200];
};

struct object
{
  int offset;
  int type;
  int ref;
  int nextoffset;
  int length;
  struct object *child;
  struct object *next, *prev;
};

struct object *objects;

void
searchobjects (int fd)
{
  unsigned int *buffer;
  int offset, rc, i;
  struct object *obj, *lastobj;
  int total;

  buffer = (unsigned int *) malloc (10240);
  obj = lastobj = NULL;
  total = 0;

  lseek (fd, 0, SEEK_SET);
  offset = 0;
  while (1)
    {
      if (offset % 1024 == 0)
	printf ("\rScanning objects ... %u MB", offset / (1024 * 1024));
      rc = read (fd, buffer, 10240);
      if (rc < 0)
	{
	  perror ("oe5dbx");
	  exit (1);
	}
      if (rc == 0)
	break;
      rc /= 4;
      for (i = 0; i < rc; i++)
	{
	  if (buffer[i] == offset + (i * 4))
	    {
	      obj = (struct object *) malloc (sizeof (struct object));
	      if (obj == NULL)
		{
		  fprintf (stderr, "Not enough memory!");
		  exit (1);
		}

	      obj->offset = offset + (i * 4);
	      obj->type = OBJ_UNKNOWN;
	      obj->prev = lastobj;
	      obj->next = NULL;
	      obj->child = NULL;
	      obj->ref = 0;
	      if (lastobj)
		lastobj->next = obj;
	      else
		objects = obj;
	      lastobj = obj;
	      total++;
	    }
	}
      offset += rc * 4;
    }
  printf ("\n Done, search found %u records\n", total);
}

void
analyzeobjects (int fd)
{
  struct object *obj, *tmpobj;
  int count, hcount;
  struct messageblock blk;

  obj = objects;
  hcount = count = 0;

  while (obj)
    {
      if (count % 1000 == 0)
	{
	  printf ("\rAnalyzing objects... %u", count);
	  fflush (stdout);
	}

      lseek (fd, obj->offset, SEEK_SET);
      if (read (fd, &blk, sizeof (blk)) != sizeof (blk))
	{
	  obj->type = OBJ_BAD;
	}
      else
	{
	  if (blk.twohundred != 0x200 || blk.msglen > 0x200)
	    {
	      obj->type = OBJ_BAD;
	      //  printf("%08lx %08lx %08lx\n", blk.twohundred, blk.msglen, blk.nextblock);
	    }
	  else
	    {
	      obj->nextoffset = blk.nextblock;
	      obj->length = blk.msglen;

	      if (blk.nextblock == 0)
		{
		  obj->type = OBJ_MSG_END;
		}
	      else
		{
		  if (blk.msglen > 13 &&
		      (!strncasecmp (blk.buffer, "Received:", 9) ||
		       !strncasecmp (blk.buffer, "Delivered-To:", 13) ||
		       !strncasecmp (blk.buffer, "Return-Path:", 12)))
		    {
		      obj->type = OBJ_MSG_BEGIN;
		      hcount++;
		      if (obj->prev)
			if (obj->prev->nextoffset == obj->offset)
			  {
			    obj->type = OBJ_MSG;
			    printf
			      ("\rBlock 0x%08lx looks like a message start, but it is NOT\n",
			       obj->offset);
			    fflush (stdout);
			    hcount--;
			  }
		    }
		  else

		    obj->type = OBJ_MSG;
		}
	    }
	}

      obj = obj->next;
      count++;
    }

  printf ("\n Done, have %u message headers\n", hcount);

}

struct object *
findobject (int offset, struct object *start)
{
  struct object *obj;

  obj =  start;

  while (obj)
    {
      if (offset == obj->offset)
	return obj;
     if (start->offset < offset)
         obj = obj->next;
         else obj = obj->prev;
    }

  return NULL;
}


void
checkobjects (int fd)
{
  struct object *obj, *tmpobj;
  int count, corruptcount;

  obj = objects;
  corruptcount = count = 0;

  while (obj)
    {
      if (count % 1000 == 0)
	{
	  printf ("\rChecking objects... %u", count);
	  fflush (stdout);
	}

      if (obj->type == OBJ_MSG || obj->type == OBJ_MSG_BEGIN)
	{
	  tmpobj = findobject (obj->nextoffset, obj);
	  if (tmpobj == NULL)
	    {
	      printf
		("\nCorrupted block 0x%08lx pointing to nonexistent object!\n",
		 obj->offset);
	      fflush (stdout);
	      obj->type = OBJ_MSG_CORRUPT;
	      corruptcount++;
	    }
	  else
	    {
	      if (tmpobj->ref)
		{
		  printf
		    ("\nCorrupted block 0x%08lx pointing to already referenced block at %08lx\n",
		     obj->offset, tmpobj->offset);
		  fflush (stdout);

		}
	      tmpobj->ref = 1;

	      if (tmpobj->type == OBJ_MSG_BEGIN || tmpobj->type == OBJ_BAD)
		{
		  printf
		    ("\nCorrupted block 0x%08lx pointing to invalid block at %08lx\n",
		     obj->offset, tmpobj->offset);
		  fflush (stdout);
		}
	    }
	}

      obj = obj->next;
      count++;
    }

  printf ("\n Done, %u corrupted messages\n", corruptcount);

}

/*
 This is terrible recursive function which links message blocks together.
 Quick & dirty non-effective solution :-((
*/
void
linkmessage (struct object *root, struct object *message)
{
  struct object *obj;

  if (message->nextoffset == 0)
    return;
  obj = findobject (message->nextoffset, root);
  if (obj == NULL || obj == root)
    return;
  if (obj->prev)
    obj->prev->next = obj->next;
  else
    objects = obj->next;
  if (obj->next)
    obj->next->prev = obj->prev;

  obj->prev = NULL;
  obj->next = NULL;
  obj->child = NULL;

  printf("%08lx: %08lx -> %08lx\n", root->offset, message->offset, obj->offset);

  message->child = obj;
  linkmessage (root, obj);
}


void
linkobjects (int fd)
{
  struct object *obj, *tmpobj;
  int count, corruptcount;

  obj = objects;
  corruptcount = count = 0;

  while (obj)
    {
      if (obj->type == OBJ_MSG_BEGIN 
//          ||
//	  (obj->type == OBJ_MSG && obj->ref == 0) ||
//	  (obj->type == OBJ_MSG_CORRUPT && obj->ref == 0) ||
//	  (obj->type == OBJ_MSG_END && obj->ref == 0)
)
	{
//      if (count % 10 == 0)
	{
	  printf ("\rLinking messages... %u", count);
	  fflush (stdout);
	}

	linkmessage (obj, obj);
      count++;
	}

      obj = obj->next;
    }

  printf ("\n Done\n");

}

void
writemessages (int fd)
{
  struct object *obj, *tmpobj;
  int count, corruptcount;
  char name[0x100];
  int msgfd;
  char buffer[0x200];

  obj = objects;
  corruptcount = count = 0;

  while (obj)
    {
      if (count % 1000 == 0)
	{
	  printf ("\rWriting messages... %u", count);
	  fflush (stdout);
	}
      if (obj->type == OBJ_MSG_BEGIN 
//	 || obj->type == OBJ_MSG ||
//	  obj->type == OBJ_MSG_CORRUPT || obj->type == OBJ_MSG_END
)
	{
	  sprintf (name, "%s%04x", obj->type == OBJ_MSG_BEGIN ? "msg" : "bad",
		   count);
	  msgfd = open (name, O_CREAT | O_RDWR, 0600);
	  if (msgfd < 0)
	    {
	      perror (name);
	    }
	  else
	    {
	      tmpobj = obj;
	      while (tmpobj)
		{
                  lseek(fd, tmpobj->offset + 0x10, SEEK_SET);
		  read (fd, buffer, tmpobj->length);
		  write (msgfd, buffer, tmpobj->length);
		  tmpobj = tmpobj->child;
		}
	      close (msgfd);
	    }
	}

      obj = obj->next;
      count++;
    }

  printf ("\n Done\n");

}


int
main (int argc,char **argv)
{
  int fd;

  fd = open (argv[1], O_RDONLY);
  if (fd < 0) {
    perror(argv[1]);
    return 1;
  }

  searchobjects (fd);
  analyzeobjects (fd);
  checkobjects (fd);
  linkobjects (fd);
  writemessages(fd);

  return 0;
}
